import os
import numpy as np
import PIL
from typing import *
from PIL import Image, ImageDraw
from omegaconf import OmegaConf

__all__ = ["DEFAULT_CONFIG", "load_configs", "generate_detection_image_with_annotation"]


def load_configs(path):

    return OmegaConf.load(path)


CONFIG_PATH = os.path.join(
    os.path.realpath("__file__"), os.pardir, "configs", "simulator.yaml"
)

DEFAULT_CONFIG = load_configs(CONFIG_PATH).DEFAULT


def generate_detection_image_with_annotation(
    src_data: Sequence[Tuple[Image.Image, int]],
    num_src: int,
    config: Optional[OmegaConf] = None,
) -> Tuple[Image.Image, List[List[int]], List[int]]:
    """
    params:
        src_data: torch Dataset class or List-like object [Image, Label]
        num_src:  number of images
        config:   simulator config

    output:
        image:  PIL Image
        boxes:  List[Box], Box is List[int] [x0, y0, x1, y1]
        labels: List[int], ground truth labels for each box
    """

    if config is None:
        config = DEFAULT_CONFIG

    canvas = config.canvas
    margin = config.margin
    stride = config.stride
    min_size = config.min_size
    max_size = config.max_size
    rotation = config.rotation
    keep_aspect = config.keep_aspect

    bgd = PIL.Image.fromarray(
        np.random.randint(100, 200, (5, 5, 3)).astype(np.uint8)
    ).resize(canvas, 3)
    bgd = PIL.ImageEnhance.Contrast(bgd).enhance(config.bgd_contrast)
    bgd = PIL.ImageEnhance.Brightness(bgd).enhance(config.bgd_bridgtness)

    xrange_ = range(0, canvas[0] - margin[0], stride[0])
    yrange_ = range(0, canvas[1] - margin[1], stride[1])
    idx_ = np.random.choice(len(src_data), num_src)
    dx_ = np.random.choice(xrange_, num_src, replace=False)
    dy_ = np.random.choice(yrange_, num_src, replace=False)
    w_ = np.random.choice(range(min_size, max_size + 1, 8), num_src)
    h_ = (
        np.random.choice(range(min_size, max_size + 1, 8), num_src)
        if not keep_aspect
        else w_
    )
    rot_ = np.random.choice(range(-rotation, rotation), num_src)

    lab_ = []
    box_ = []
    img_ = []
    for jdx, idx in enumerate(idx_):

        img, lab = src_data[idx]

        img = img.resize((w_[jdx], h_[jdx]), resample=3)
        img = img.rotate(rot_[jdx], resample=3, expand=True)
        img_.append(img)

        box = list(img.getbbox())

        dx, dy = dx_[jdx], dy_[jdx]

        box[0::2] += dx
        box[1::2] += dy
        box[0] = clip(box[0], 0, canvas[0])
        box[2] = clip(box[2], 0, canvas[0])
        box[1] = clip(box[1], 0, canvas[1])
        box[3] = clip(box[3], 0, canvas[1])

        if (box[2] - box[0]) * (box[3] - box[1]) > 0:
            lab_.append(lab)
            box_.append(box)
            color = np.random.randint(100, 200, (3,))
            fgd = PIL.ImageOps.colorize(img, black=color * 0.7, white=color).convert(
                "RGB"
            )
            fgd = PIL.ImageEnhance.Contrast(fgd).enhance(config.fgd_contrast)
            fgd = PIL.ImageEnhance.Brightness(fgd).enhance(config.fgd_bridgtness)
            msk = img.filter(PIL.ImageFilter.GaussianBlur(radius=config.blur_radius))
            bgd.paste(fgd, (dx, dy), msk)

    return bgd, box_, lab_


def clip(x, minval, maxval):
    if x < minval:
        return minval
    if x > maxval:
        return maxval
    return x



from enum import Enum
from typing import *
from PIL import Image
import torch
import torchvision.transforms.functional as TF


__all__ = ["CollateMode", "CollateFunctionBase", "collate_boxes", "collate_pil_images"]


class CollateMode(Enum):

    MOSAIC = "mosaic"
    TOLIST = "tolist"


def collate_pil_images(
    images: Union[Sequence[Image.Image], Sequence[torch.Tensor]],
    rows: int,
    columns: int,
    atom_size: Optional[Tuple[int, int]] = None,
    mode: str = "RGB",
    to_tensor: bool = False,
) -> Union[Image.Image, torch.Tensor]:

    if atom_size is None:
        atom_size = images[0].size
    else:
        atom_size = tuple(atom_size)

    pil_images = []
    for img in images:
        if isinstance(img, torch.Tensor):
            pil_img = TF.to_pil_img(img)
        elif isinstance(img, Image.Image):
            pil_img = img
        else:
            raise TypeError(
                f"Type of image {type(img)} is not in (`PIL.Image.Image`, `torch.Tensor`)"
            )
        assert pil_img.size == atom_size
        pil_images.append(pil_img)

    w, h = atom_size

    assert rows * columns == len(images)

    W, H = w * columns, h * rows

    img = Image.new(mode, (W, H))
    idx = 0
    for rdx in range(rows):
        for cdx in range(columns):

            img.paste(images[idx], (w * cdx, h * rdx))
            idx += 1

    return TF.to_tensor(img) if to_tensor else img


def collate_boxes(
    boxes: List[torch.Tensor],  # boxes in relative coordinates, float tensors
    rows: int,
    columns: int,
) -> torch.Tensor:

    idx = 0
    for rdx in range(rows):
        for cdx in range(columns):
            boxes[idx][:, 0::2] += cdx
            boxes[idx][:, 1::2] += rdx
            idx += 1

    return torch.cat(boxes)


class CollateFunctionBase(Callable):
    """CollateFunctionBase:
    factory class of collate functions

    params:
        keys: sequence of keywords
        mode: collation mode
        mosaic_row_col: number of rows and number of columns of mosaic collation

    example usage:

        collate_fn = CollateFunctionBase(keys=['image', 'bbox', 'label'], mode='tolist')

        or

        collate_fn = CollateFunctionBase(keys=['image', 'mask', 'bbox', 'label'], mode='mosaic', mosaic_row_col=(3, 3))
    """

    REPR_INDENT = 2
    KEYS = {"image", "mask", "bbox", "label"}

    def __init__(
        self,
        keys: Sequence[str],
        mode: Union[CollateMode, str],
        mosaic_row_col: Optional[Tuple[int, int]] = None,
    ) -> None:

        for k in keys:
            assert k in self.KEYS
        self.keys = keys
        if isinstance(mode, str):
            mode = CollateMode(mode)
        assert isinstance(mode, CollateMode)
        self.mode = mode
        if mode == CollateMode.MOSAIC:
            assert mosaic_row_col is not None
        self._row_col = mosaic_row_col

    def __call__(self, batch: Dict[str, Tuple[Any]]) -> Tuple[Any]:

        outputs = {k: [] for k in self.keys}
        for item in batch:
            for k, v in item.items():
                outputs[k].append(v)

        if self.mode == CollateMode.TOLIST:
            return tuple([outputs[k] for k in self.keys])

        elif self.mode == CollateMode.MOSAIC:
            row, col = self._row_col
            for k in self.keys:
                if k == "image" or k == "mask":
                    outputs[k] = collate_pil_images(outputs[k], rows=row, columns=col)
                elif k == "bbox":
                    outputs[k] = collate_boxes(outputs[k], rows=row, columns=col)
                elif k == "label":
                    outputs[k] = torch.cat(outputs[k], 0)
                else:
                    raise NotImplementedError(f"Keyword: {k} is not implemented.")

            return tuple([outputs[k] for k in self.keys])
        else:
            raise NotImplementedError(f"Collate mode: {self.mode} is not implemented.")

    def __repr__(self) -> str:
        out = []
        out.append(self.__class__.__name__)
        for k, v in self.__dict__.items():
            out.append(self.REPR_INDENT * " " + f"{k}: {v}")

        return "\n".join(out)



from setuptools import find_packages, setup

# load readme
with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="mnist-detection",
    version="0.1.1",
    author="Chenchao Zhao",
    author_email="chenchao.zhao@gmail.com",
    description="Simulated MNIST detection dataset",
    packages=find_packages(exclude=["tests", "scripts"]),
    package_data={'configs': ['configs']},
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["torch", "numpy", "pillow", "omegaconf", "torchvision"],
    license="GNU",
)
